/* Kernel module to check whether IP and MAC address are listed as a pair
   in the specified table.
   
   Author: Chris Wilson <chris@netservers.co.uk>
   Copyright 2004 Netservers Ltd. [http://www.netservers.co.uk]

   Based on ipt_recent.c:
   Copyright 2002-2003, Stephen Frost
   Author: Stephen Frost <sfrost@snowman.net>
   Project Page: http://snowman.net/projects/ipt_recent/

   This software is distributed under the terms of the GPL, Version 2
   This copyright does not cover user programs that use kernel services
   by normal system calls.
*/

#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#include <iptables.h>
#include <linux/netfilter_ipv4/ipt_macmatch.h>

/* Function which prints out usage message. */
static void help ()
{
	printf(MACMATCH_NAME " " MACMATCH_VER " options:\n"
               " --mac\n"
               "\t\t\t\tMatch if the MAC address is in the table\n"
               " --ip\n"
               "\t\t\t\tMatch if the IP address is in the table\n"
               " --mac --ip\n"
               "\t\t\t\tMatch if the MAC and IP exist as a pair in table\n"
               "\n"
               "One or both of --mac and --ip is required. The interface is always matched.\n");
}
  
static struct option opts[] = {
	{ "mac", 0, 0, 201 }, 
	{ "ip",  0, 0, 202 }, 
	{0}
};

/* Initialize the match. */
static void init (struct ipt_entry_match *match, unsigned int *nfcache)
{
	mm_info_t *info;
	*nfcache |= NFC_UNKNOWN;
        info = (mm_info_t *)(match)->data;
	info->flags = 0;
}

/* Function which parses command options; returns true if it ate an option */
static int parse (int c, char **argv, int invert, unsigned int *flags,
                  const struct ipt_entry *entry,
                  unsigned int *nfcache,
                  struct ipt_entry_match **match)
{
	mm_info_t *info = (mm_info_t *)(*match)->data;

	switch (c) {
        case 201:
                info->flags |= MM_FLAG_MATCHMAC;
                *flags = 1;
                break;
			
        case 202:
                info->flags |= MM_FLAG_MATCHIP;
                *flags = 1;
                break;

        default:
                return 0;
        }

	return 1;
}

/* Final check; must have specified a specific option. */
static void final_check (unsigned int flags)
{
        if (!flags) {
		exit_error(PARAMETER_PROBLEM,
                           "macmatch: you must specify either --mac or "
                           "--ip or both");
        }
}

/* Prints out the matchinfo. */
static void print (const struct ipt_ip *ip,
                   const struct ipt_entry_match *match,
                   int numeric)
{
        mm_info_t *info = (mm_info_t *)(match->data);
        
        printf("macmatch: match ");

        if (info->flags & MM_FLAG_MATCHMAC) {
                printf("mac");
        }

        if ( (info->flags & MM_FLAG_MATCHMAC) &&
             (info->flags & MM_FLAG_MATCHIP) ) {
                printf(" and ");
        }

        if (info->flags & MM_FLAG_MATCHIP) {
                printf("ip");
        }
}

/* Saves the union ipt_matchinfo in parsable form to stdout. */
static void save (const struct ipt_ip *ip, 
                  const struct ipt_entry_match *match)
{
        mm_info_t *info = (mm_info_t *)(match->data);

        printf("-m macmatch");

        if (info->flags & MM_FLAG_MATCHMAC) {
                printf("--mac");
        }

        if (info->flags & MM_FLAG_MATCHMAC) {
                printf("--ip");
        }
}

static
struct iptables_match mm_match
= { NULL,
    "macmatch",
    IPTABLES_VERSION,
    IPT_ALIGN(sizeof(mm_info_t)),
    IPT_ALIGN(sizeof(mm_info_t)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_match(&mm_match);
}
